# """
#     event.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

from typing import Optional, List
from snowplow_tracker import payload
from snowplow_tracker.subject import Subject

from snowplow_tracker.self_describing_json import SelfDescribingJson

from snowplow_tracker.constants import CONTEXT_SCHEMA
from snowplow_tracker.typing import JsonEncoderFunction, PayloadDict


class Event(object):
    """
    Event class which contains
    elements that can be set in all events. These are context, trueTimestamp, and Subject.

    Context is a list of custom SelfDescribingJson entities.
    TrueTimestamp is a user-defined timestamp.
    Subject is an event-specific Subject. Its fields will override those of the
    Tracker-associated Subject, if present.

    """

    def __init__(
        self,
        dict_: Optional[PayloadDict] = None,
        event_subject: Optional[Subject] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        true_timestamp: Optional[float] = None,
    ) -> None:
        """
        Constructor
        :param  dict_:              Optional Dictionary to be added to the Events Payload
        :type   dict_:              dict(string:\\*) | None
        :param  event_subject:      Optional per event subject
        :type   event_subject:      subject | None
        :param  context:            Custom context for the event
        :type   context:            context_array | None
        :param  true_timestamp:     Optional event timestamp in milliseconds
        :type   true_timestamp:     int | float | None

        """
        self.payload = payload.Payload(dict_=dict_)
        self.event_subject = event_subject
        self.context = context or []
        self.true_timestamp = true_timestamp

    def build_payload(
        self,
        encode_base64: bool,
        json_encoder: Optional[JsonEncoderFunction],
        subject: Optional[Subject] = None,
    ) -> "payload.Payload":
        """
        :param encode_base64:    Whether JSONs in the payload should be base-64 encoded
        :type  encode_base64:    bool
        :param json_encoder:     Custom JSON serializer that gets called on non-serializable object
        :type  json_encoder:     function | None
        :param  subject:         Optional per event subject
        :type   subject:         subject | None
        :rtype:                  payload.Payload
        """
        if len(self.context) > 0:
            context_jsons = list(map(lambda c: c.to_json(), self.context))
            context_envelope = SelfDescribingJson(
                CONTEXT_SCHEMA, context_jsons
            ).to_json()
            self.payload.add_json(
                context_envelope, encode_base64, "cx", "co", json_encoder
            )

        if isinstance(
            self.true_timestamp,
            (
                int,
                float,
            ),
        ):
            self.payload.add("ttm", int(self.true_timestamp))

        if self.event_subject is not None:
            fin_payload_dict = self.event_subject.combine_subject(subject)
        else:
            fin_payload_dict = {} if subject is None else subject.standard_nv_pairs

        self.payload.add_dict(fin_payload_dict)
        return self.payload

    @property
    def event_subject(self) -> Optional[Subject]:
        """
        Optional per event subject
        """
        return self._event_subject

    @event_subject.setter
    def event_subject(self, value: Optional[Subject]):
        self._event_subject = value

    @property
    def context(self) -> List[SelfDescribingJson]:
        """
        Custom context for the event
        """
        return self._context

    @context.setter
    def context(self, value: List[SelfDescribingJson]):
        self._context = value

    @property
    def true_timestamp(self) -> Optional[float]:
        """
        Optional event timestamp in milliseconds
        """
        return self._true_timestamp

    @true_timestamp.setter
    def true_timestamp(self, value: Optional[float]):
        self._true_timestamp = value



# """
#     self_describing_json.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import json
from typing import Union

from snowplow_tracker.typing import PayloadDict, PayloadDictList
from snowplow_tracker.contracts import non_empty_string


class SelfDescribingJson(object):
    def __init__(self, schema: str, data: Union[PayloadDict, PayloadDictList]) -> None:
        self.schema = schema
        self.data = data

    @property
    def schema(self) -> str:
        return self._schema

    @schema.setter
    def schema(self, value: str):
        non_empty_string(value)
        self._schema = value

    def to_json(self) -> PayloadDict:
        return {"schema": self.schema, "data": self.data}

    def to_string(self) -> str:
        return json.dumps(self.to_json())



# """
#     screen_view.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

from typing import Dict, Optional, List
from snowplow_tracker.typing import JsonEncoderFunction
from snowplow_tracker.events.event import Event
from snowplow_tracker.events.self_describing import SelfDescribing
from snowplow_tracker import SelfDescribingJson
from snowplow_tracker.constants import (
    MOBILE_SCHEMA_PATH,
    SCHEMA_TAG,
)
from snowplow_tracker import payload
from snowplow_tracker.subject import Subject
from snowplow_tracker.contracts import non_empty_string


class ScreenView(Event):
    """
    Constructs a ScreenView event object.

    When tracked, generates a SelfDescribing event (event type "ue").

    Schema: `iglu:com.snowplowanalytics.mobile/screen_view/jsonschema/1-0-0`
    """

    def __init__(
        self,
        id_: str,
        name: str,
        type: Optional[str] = None,
        previous_name: Optional[str] = None,
        previous_id: Optional[str] = None,
        previous_type: Optional[str] = None,
        transition_type: Optional[str] = None,
        event_subject: Optional[Subject] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        true_timestamp: Optional[float] = None,
    ) -> None:
        """
        :param  id_:            Screen view ID. This must be of type UUID.
        :type   id_:            string
        :param  name:           The name of the screen view event
        :type   name:           string
        :param  type:           The type of screen that was viewed e.g feed / carousel.
        :type   type:           string | None
        :param  previous_name:  The name of the previous screen.
        :type   previous_name:  string | None
        :param  previous_id:    The screenview ID of the previous screenview.
        :type   previous_id:    string | None
        :param  previous_type   The screen type of the previous screenview
        :type   previous_type   string | None
        :param  transition_type The type of transition that led to the screen being viewed.
        :type   transition_type string | None
        :param  event_subject:   Optional per event subject
        :type   event_subject:   subject | None
        :param  context:         Custom context for the event
        :type   context:         context_array | None
        :param  true_timestamp:  Optional event timestamp in milliseconds
        :type   true_timestamp:  int | float | None
        """
        super(ScreenView, self).__init__(
            event_subject=event_subject, context=context, true_timestamp=true_timestamp
        )
        self.screen_view_properties: Dict[str, str] = {}
        self.id_ = id_
        self.name = name
        self.type = type
        self.previous_name = previous_name
        self.previous_id = previous_id
        self.previous_type = previous_type
        self.transition_type = transition_type

    @property
    def id_(self) -> str:
        """
        Screen view ID. This must be of type UUID.
        """
        return self.screen_view_properties["id"]

    @id_.setter
    def id_(self, value: str):
        non_empty_string(value)
        self.screen_view_properties["id"] = value

    @property
    def name(self) -> str:
        """
        The name of the screen view event
        """
        return self.screen_view_properties["name"]

    @name.setter
    def name(self, value: str):
        non_empty_string(value)
        self.screen_view_properties["name"] = value

    @property
    def type(self) -> Optional[str]:
        """
        The type of screen that was viewed e.g feed / carousel
        """
        return self.screen_view_properties["type"]

    @type.setter
    def type(self, value: Optional[str]):
        if value is not None:
            self.screen_view_properties["type"] = value

    @property
    def previous_name(self) -> Optional[str]:
        """
        The name of the previous screen.
        """
        return self.screen_view_properties["previousName"]

    @previous_name.setter
    def previous_name(self, value: Optional[str]):
        if value is not None:
            self.screen_view_properties["previousName"] = value

    @property
    def previous_id(self) -> Optional[str]:
        """
        The screenview ID of the previous screenview.
        """
        return self.screen_view_properties["previousId"]

    @previous_id.setter
    def previous_id(self, value: Optional[str]):
        if value is not None:
            self.screen_view_properties["previousId"] = value

    @property
    def previous_type(self) -> Optional[str]:
        """
        The screen type of the previous screenview
        """
        return self.screen_view_properties["previousType"]

    @previous_type.setter
    def previous_type(self, value: Optional[str]):
        if value is not None:
            self.screen_view_properties["previousType"] = value

    @property
    def transition_type(self) -> Optional[str]:
        """
        The type of transition that led to the screen being viewed
        """
        return self.screen_view_properties["transitionType"]

    @transition_type.setter
    def transition_type(self, value: Optional[str]):
        if value is not None:
            self.screen_view_properties["transitionType"] = value

    def build_payload(
        self,
        encode_base64: bool,
        json_encoder: Optional[JsonEncoderFunction],
        subject: Optional[Subject] = None,
    ) -> "payload.Payload":
        """
        :param encode_base64:    Whether JSONs in the payload should be base-64 encoded
        :type  encode_base64:    bool
        :param json_encoder:     Custom JSON serializer that gets called on non-serializable object
        :type  json_encoder:     function | None
        :param  subject:         Optional per event subject
        :type   subject:         subject | None
        :rtype:                  payload.Payload
        """
        event_json = SelfDescribingJson(
            "%s/screen_view/%s/1-0-0" % (MOBILE_SCHEMA_PATH, SCHEMA_TAG),
            self.screen_view_properties,
        )
        self_describing = SelfDescribing(
            event_json=event_json,
            event_subject=self.event_subject,
            context=self.context,
            true_timestamp=self.true_timestamp,
        )
        return self_describing.build_payload(
            encode_base64, json_encoder, subject=subject
        )



# """
#     constants.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """
from typing import List
from snowplow_tracker import _version, SelfDescribingJson

VERSION = "py-%s" % _version.__version__
DEFAULT_ENCODE_BASE64: bool = True  # Type hint required for Python 3.6 MyPy check
BASE_SCHEMA_PATH = "iglu:com.snowplowanalytics.snowplow"
MOBILE_SCHEMA_PATH = "iglu:com.snowplowanalytics.mobile"
SCHEMA_TAG = "jsonschema"
CONTEXT_SCHEMA = "%s/contexts/%s/1-0-1" % (BASE_SCHEMA_PATH, SCHEMA_TAG)
UNSTRUCT_EVENT_SCHEMA = "%s/unstruct_event/%s/1-0-0" % (BASE_SCHEMA_PATH, SCHEMA_TAG)
ContextArray = List[SelfDescribingJson]



# """
#     self_describing.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """
from typing import Optional, List
from snowplow_tracker.typing import JsonEncoderFunction
from snowplow_tracker.events.event import Event
from snowplow_tracker import SelfDescribingJson
from snowplow_tracker.constants import UNSTRUCT_EVENT_SCHEMA
from snowplow_tracker import payload
from snowplow_tracker.subject import Subject
from snowplow_tracker.contracts import non_empty


class SelfDescribing(Event):
    """
    Constructs a SelfDescribing event object.

    This is a customisable event type which allows you to track anything describable
    by a JsonSchema.

    When tracked, generates a self-describing event (event type "ue").
    """

    def __init__(
        self,
        event_json: SelfDescribingJson,
        event_subject: Optional[Subject] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        true_timestamp: Optional[float] = None,
    ) -> None:
        """
        :param  event_json:      The properties of the event. Has two field:
                                 A "data" field containing the event properties and
                                 A "schema" field identifying the schema against which the data is validated
        :type   event_json:      self_describing_json
        :param  event_subject:   Optional per event subject
        :type   event_subject:   subject | None
        :param  context:         Custom context for the event
        :type   context:         context_array | None
        :param  true_timestamp:          Optional event timestamp in milliseconds
        :type   true_timestamp:          int | float | None
        """
        super(SelfDescribing, self).__init__(
            event_subject=event_subject, context=context, true_timestamp=true_timestamp
        )
        self.payload.add("e", "ue")
        self.event_json = event_json

    @property
    def event_json(self) -> SelfDescribingJson:
        """
        The properties of the event. Has two field:
            A "data" field containing the event properties and
            A "schema" field identifying the schema against which the data is validated
        """
        return self._event_json

    @event_json.setter
    def event_json(self, value: SelfDescribingJson):
        self._event_json = value

    def build_payload(
        self,
        encode_base64: bool,
        json_encoder: Optional[JsonEncoderFunction],
        subject: Optional[Subject] = None,
    ) -> "payload.Payload":
        """
        :param encode_base64:    Whether JSONs in the payload should be base-64 encoded
        :type  encode_base64:    bool
        :param json_encoder:     Custom JSON serializer that gets called on non-serializable object
        :type  json_encoder:     function | None
        :param  subject:         Optional per event subject
        :type   subject:         subject | None
        :rtype:                  payload.Payload
        """

        envelope = SelfDescribingJson(
            UNSTRUCT_EVENT_SCHEMA, self.event_json.to_json()
        ).to_json()
        self.payload.add_json(envelope, encode_base64, "ue_px", "ue_pr", json_encoder)

        return super(SelfDescribing, self).build_payload(
            encode_base64=encode_base64, json_encoder=json_encoder, subject=subject
        )



# """
#     emitters.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import logging
import time
import threading
import requests
import random
from typing import Optional, Union, Tuple, Dict, cast, Callable
from queue import Queue

from snowplow_tracker.self_describing_json import SelfDescribingJson
from snowplow_tracker.typing import (
    PayloadDict,
    PayloadDictList,
    HttpProtocol,
    Method,
    SuccessCallback,
    FailureCallback,
    EmitterProtocol,
)
from snowplow_tracker.contracts import one_of
from snowplow_tracker.event_store import EventStore, InMemoryEventStore

# logging
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

DEFAULT_MAX_LENGTH = 10
PAYLOAD_DATA_SCHEMA = (
    "iglu:com.snowplowanalytics.snowplow/payload_data/jsonschema/1-0-4"
)
PROTOCOLS = {"http", "https"}
METHODS = {"get", "post"}


# Unifes the two request methods under one interface
class Requester:
    post: Callable
    get: Callable

    def __init__(self, post: Callable, get: Callable):
        # 3.6 MyPy compatibility:
        # error: Cannot assign to a method
        # https://github.com/python/mypy/issues/2427
        setattr(self, "post", post)
        setattr(self, "get", get)


class Emitter(EmitterProtocol):
    """
    Synchronously send Snowplow events to a Snowplow collector
    Supports both GET and POST requests
    """

    def __init__(
        self,
        endpoint: str,
        protocol: HttpProtocol = "https",
        port: Optional[int] = None,
        method: Method = "post",
        batch_size: Optional[int] = None,
        on_success: Optional[SuccessCallback] = None,
        on_failure: Optional[FailureCallback] = None,
        byte_limit: Optional[int] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        max_retry_delay_seconds: int = 60,
        buffer_capacity: Optional[int] = None,
        custom_retry_codes: Dict[int, bool] = {},
        event_store: Optional[EventStore] = None,
        session: Optional[requests.Session] = None,
    ) -> None:
        """
        :param endpoint:    The collector URL. If protocol is not set in endpoint it will automatically set to "https://" - this is done automatically.
        :type  endpoint:    string
        :param protocol:    The protocol to use - http or https. Defaults to https.
        :type  protocol:    protocol
        :param port:        The collector port to connect to
        :type  port:        int | None
        :param method:      The HTTP request method. Defaults to post.
        :type  method:      method
        :param batch_size:  The maximum number of queued events before the buffer is flushed. Default is 10.
        :type  batch_size:  int | None
        :param on_success:  Callback executed after every HTTP request in a flush has status code 200
                            Gets passed one argument, an array of dictionaries corresponding to the sent events' payloads
        :type  on_success:  function | None
        :param on_failure:  Callback executed if at least one HTTP request in a flush has status code other than 200
                            Gets passed two arguments:
                            1) The number of events which were successfully sent
                            2) An array of dictionaries corresponding to the unsent events' payloads
        :type  on_failure:  function | None
        :param byte_limit:  The size event list after reaching which queued events will be flushed
        :type  byte_limit:  int | None
        :param request_timeout: Timeout for the HTTP requests. Can be set either as single float value which
                                 applies to both "connect" AND "read" timeout, or as tuple with two float values
                                 which specify the "connect" and "read" timeouts separately
        :type request_timeout:  float | tuple | None
        :param max_retry_delay_seconds:     Set the maximum time between attempts to send failed events to the collector. Default 60 seconds
        :type max_retry_delay_seconds:      int
        :param buffer_capacity: The maximum capacity of the event buffer.
                                When the buffer is full new events are lost.
        :type buffer_capacity: int
        :param  custom_retry_codes: Set custom retry rules for HTTP status codes received in emit responses from the Collector.
                                    By default, retry will not occur for status codes 400, 401, 403, 410 or 422. This can be overridden here.
                                    Note that 2xx codes will never retry as they are considered successful.
        :type   custom_retry_codes: dict
        :param  event_store:    Stores the event buffer and buffer capacity. Default is an InMemoryEventStore object with buffer_capacity of 10,000 events.
        :type   event_store:    EventStore | None
        :param  session:    Persist parameters across requests by using a session object
        :type   session:    requests.Session | None
        """
        one_of(protocol, PROTOCOLS)
        one_of(method, METHODS)

        self.endpoint = Emitter.as_collector_uri(endpoint, protocol, port, method)

        self.method = method

        if event_store is None:
            if buffer_capacity is None:
                event_store = InMemoryEventStore(logger=logger)
            else:
                event_store = InMemoryEventStore(
                    buffer_capacity=buffer_capacity, logger=logger
                )

        self.event_store = event_store

        if batch_size is None:
            if method == "post":
                batch_size = DEFAULT_MAX_LENGTH
            else:
                batch_size = 1

        if buffer_capacity is not None and batch_size > buffer_capacity:
            batch_size = buffer_capacity

        self.batch_size = batch_size
        self.byte_limit = byte_limit
        self.bytes_queued = None if byte_limit is None else 0
        self.request_timeout = request_timeout

        self.on_success = on_success
        self.on_failure = on_failure

        self.lock = threading.RLock()

        self.timer = FlushTimer(emitter=self, repeating=True)
        self.retry_timer = FlushTimer(emitter=self, repeating=False)

        self.max_retry_delay_seconds = max_retry_delay_seconds
        self.retry_delay: Union[int, float] = 0

        self.custom_retry_codes = custom_retry_codes
        logger.info("Emitter initialized with endpoint " + self.endpoint)

        if session is None:
            self.request_method = Requester(post=requests.post, get=requests.get)
        else:
            self.request_method = Requester(post=session.post, get=session.get)

    @staticmethod
    def as_collector_uri(
        endpoint: str,
        protocol: HttpProtocol = "https",
        port: Optional[int] = None,
        method: Method = "post",
    ) -> str:
        """
        :param endpoint:  The raw endpoint provided by the user
        :type  endpoint:  string
        :param protocol:  The protocol to use - http or https
        :type  protocol:  protocol
        :param port:      The collector port to connect to
        :type  port:      int | None
        :param method:    Either `get` or `post` HTTP method
        :type  method:    method
        :rtype:           string
        """
        if len(endpoint) < 1:
            raise ValueError("No endpoint provided.")

        endpoint = endpoint.rstrip("/")

        if endpoint.split("://")[0] in PROTOCOLS:
            endpoint_arr = endpoint.split("://")
            protocol = cast(HttpProtocol, endpoint_arr[0])
            endpoint = endpoint_arr[1]

        if method == "get":
            path = "/i"
        else:
            path = "/com.snowplowanalytics.snowplow/tp2"
        if port is None:
            return protocol + "://" + endpoint + path
        else:
            return protocol + "://" + endpoint + ":" + str(port) + path

    def input(self, payload: PayloadDict) -> None:
        """
        Adds an event to the buffer.
        If the maximum size has been reached, flushes the buffer.

        :param payload:   The name-value pairs for the event
        :type  payload:   dict(string:\\*)
        """
        with self.lock:
            if self.bytes_queued is not None:
                self.bytes_queued += len(str(payload))

            if self.method == "post":
                self.event_store.add_event({key: str(payload[key]) for key in payload})
            else:
                self.event_store.add_event(payload)

            if self.reached_limit():
                self.flush()

    def reached_limit(self) -> bool:
        """
        Checks if event-size or bytes limit are reached

        :rtype: bool
        """
        if self.byte_limit is None:
            return self.event_store.size() >= self.batch_size
        else:
            return (
                self.bytes_queued or 0
            ) >= self.byte_limit or self.event_store.size() >= self.batch_size

    def flush(self) -> None:
        """
        Sends all events in the buffer to the collector.
        """
        with self.lock:
            if self.retry_timer.is_active():
                return
            send_events = self.event_store.get_events_batch()
            self.send_events(send_events)
            if self.bytes_queued is not None:
                self.bytes_queued = 0

    def http_post(self, data: str) -> int:
        """
        :param data:  The array of JSONs to be sent
        :type  data:  string
        """
        logger.info("Sending POST request to %s..." % self.endpoint)
        logger.debug("Payload: %s" % data)
        try:
            r = self.request_method.post(
                self.endpoint,
                data=data,
                headers={"Content-Type": "application/json; charset=utf-8"},
                timeout=self.request_timeout,
            )
        except requests.RequestException as e:
            logger.warning(e)
            return -1

        return r.status_code

    def http_get(self, payload: PayloadDict) -> int:
        """
        :param payload:  The event properties
        :type  payload:  dict(string:\\*)
        """
        logger.info("Sending GET request to %s..." % self.endpoint)
        logger.debug("Payload: %s" % payload)
        try:
            r = self.request_method.get(
                self.endpoint, params=payload, timeout=self.request_timeout
            )
        except requests.RequestException as e:
            logger.warning(e)
            return -1

        return r.status_code

    def sync_flush(self) -> None:
        """
        Calls the flush method of the base Emitter class.
        This is guaranteed to be blocking, not asynchronous.
        """
        logger.debug("Starting synchronous flush...")
        self.flush()
        logger.info("Finished synchronous flush")

    @staticmethod
    def is_good_status_code(status_code: int) -> bool:
        """
        :param status_code:  HTTP status code
        :type  status_code:  int
        :rtype:              bool
        """
        return 200 <= status_code < 300

    def send_events(self, evts: PayloadDictList) -> None:
        """
        :param evts: Array of events to be sent
        :type  evts: list(dict(string:\\*))
        """
        if len(evts) > 0:
            logger.info("Attempting to send %s events" % len(evts))

            Emitter.attach_sent_timestamp(evts)
            success_events = []
            failure_events = []

            if self.method == "post":
                data = SelfDescribingJson(PAYLOAD_DATA_SCHEMA, evts).to_string()
                status_code = self.http_post(data)
                request_succeeded = Emitter.is_good_status_code(status_code)
                if request_succeeded:
                    success_events += evts
                else:
                    failure_events += evts

            elif self.method == "get":
                for evt in evts:
                    status_code = self.http_get(evt)
                    request_succeeded = Emitter.is_good_status_code(status_code)

                    if request_succeeded:
                        success_events += [evt]
                    else:
                        failure_events += [evt]

            if self.on_success is not None and len(success_events) > 0:
                self.on_success(success_events)
            if self.on_failure is not None and len(failure_events) > 0:
                self.on_failure(len(success_events), failure_events)

            if self._should_retry(status_code):
                self._set_retry_delay()
                self._retry_failed_events(failure_events)
            else:
                self.event_store.cleanup(success_events, False)
                self._reset_retry_delay()
        else:
            logger.info("Skipping flush since buffer is empty")

    def _set_retry_timer(self, timeout: float) -> None:
        """
        Set an interval at which failed events will be retried

        :param timeout:   interval in seconds
        :type  timeout:   int | float
        """
        self.retry_timer.start(timeout=timeout)

    def set_flush_timer(self, timeout: float) -> None:
        """
        Set an interval at which the buffer will be flushed
        :param timeout:   interval in seconds
        :type  timeout:   int | float
        """
        self.timer.start(timeout=timeout)

    def cancel_flush_timer(self) -> None:
        """
        Abort automatic async flushing
        """
        self.timer.cancel()

    @staticmethod
    def attach_sent_timestamp(events: PayloadDictList) -> None:
        """
        Attach (by mutating in-place) current timestamp in milliseconds
        as `stm` param

        :param events: Array of events to be sent
        :type  events: list(dict(string:\\*))
        :rtype: None
        """

        def update(e: PayloadDict) -> None:
            e.update({"stm": str(int(time.time()) * 1000)})

        for event in events:
            update(event)

    def _should_retry(self, status_code: int) -> bool:
        """
        Checks if a request should be retried

        :param  status_code: Response status code
        :type   status_code: int
        :rtype: bool
        """
        if Emitter.is_good_status_code(status_code):
            return False

        if status_code in self.custom_retry_codes.keys():
            return self.custom_retry_codes[status_code]

        return status_code not in [400, 401, 403, 410, 422]

    def _set_retry_delay(self) -> None:
        """
        Sets a delay to retry failed events
        """
        random_noise = random.random()
        self.retry_delay = min(
            self.retry_delay * 2 + random_noise, self.max_retry_delay_seconds
        )

    def _reset_retry_delay(self) -> None:
        """
        Resets retry delay to 0
        """
        self.retry_delay = 0

    def _retry_failed_events(self, failed_events) -> None:
        """
        Adds failed events back to the buffer to retry

        :param  failed_events: List of failed events
        :type   List
        """
        self.event_store.cleanup(failed_events, True)
        self._set_retry_timer(self.retry_delay)

    def _cancel_retry_timer(self) -> None:
        """
        Cancels a retry timer
        """
        self.retry_timer.cancel()

    # This is only here to satisfy the `EmitterProtocol` interface
    def async_flush(self) -> None:
        return


class AsyncEmitter(Emitter):
    """
    Uses threads to send HTTP requests asynchronously
    """

    def __init__(
        self,
        endpoint: str,
        protocol: HttpProtocol = "http",
        port: Optional[int] = None,
        method: Method = "post",
        batch_size: Optional[int] = None,
        on_success: Optional[SuccessCallback] = None,
        on_failure: Optional[FailureCallback] = None,
        thread_count: int = 1,
        byte_limit: Optional[int] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        max_retry_delay_seconds: int = 60,
        buffer_capacity: Optional[int] = None,
        custom_retry_codes: Dict[int, bool] = {},
        event_store: Optional[EventStore] = None,
        session: Optional[requests.Session] = None,
    ) -> None:
        """
        :param endpoint:    The collector URL. If protocol is not set in endpoint it will automatically set to "https://" - this is done automatically.
        :type  endpoint:    string
        :param protocol:    The protocol to use - http or https. Defaults to http.
        :type  protocol:    protocol
        :param port:        The collector port to connect to
        :type  port:        int | None
        :param method:      The HTTP request method
        :type  method:      method
        :param batch_size: The maximum number of queued events before the buffer is flushed. Default is 10.
        :type  batch_size: int | None
        :param on_success:  Callback executed after every HTTP request in a flush has status code 200
                            Gets passed one argument, an array of dictionaries corresponding to the sent events' payloads
        :type  on_success:  function | None
        :param on_failure:  Callback executed if at least one HTTP request in a flush has status code other than 200
                            Gets passed two arguments:
                            1) The number of events which were successfully sent
                            2) An array of dictionaries corresponding to the unsent events' payloads
        :type  on_failure:  function | None
        :param thread_count: Number of worker threads to use for HTTP requests
        :type  thread_count: int
        :param byte_limit:  The size event list after reaching which queued events will be flushed
        :type  byte_limit:  int | None
        :param max_retry_delay_seconds:     Set the maximum time between attempts to send failed events to the collector. Default 60 seconds
        :type max_retry_delay_seconds:      int
        :param buffer_capacity: The maximum capacity of the event buffer.
                                When the buffer is full new events are lost.
        :type buffer_capacity: int
        :param  event_store:    Stores the event buffer and buffer capacity. Default is an InMemoryEventStore object with buffer_capacity of 10,000 events.
        :type   event_store:    EventStore
        :param  session:    Persist parameters across requests by using a session object
        :type   session:    requests.Session | None
        """
        super(AsyncEmitter, self).__init__(
            endpoint=endpoint,
            protocol=protocol,
            port=port,
            method=method,
            batch_size=batch_size,
            on_success=on_success,
            on_failure=on_failure,
            byte_limit=byte_limit,
            request_timeout=request_timeout,
            max_retry_delay_seconds=max_retry_delay_seconds,
            buffer_capacity=buffer_capacity,
            custom_retry_codes=custom_retry_codes,
            event_store=event_store,
            session=session,
        )
        self.queue: Queue = Queue()
        for i in range(thread_count):
            t = threading.Thread(target=self.consume)
            t.daemon = True
            t.start()

    def sync_flush(self) -> None:
        while True:
            self.flush()
            self.queue.join()
            if self.event_store.size() < 1:
                break

    def flush(self) -> None:
        """
        Removes all dead threads, then creates a new thread which
        executes the flush method of the base Emitter class
        """
        with self.lock:
            self.queue.put(self.event_store.get_events_batch())
            if self.bytes_queued is not None:
                self.bytes_queued = 0

    def consume(self) -> None:
        while True:
            evts = self.queue.get()
            self.send_events(evts)
            self.queue.task_done()


class FlushTimer(object):
    """
    Internal class used by the Emitter to schedule flush calls for later.
    """

    def __init__(self, emitter: Emitter, repeating: bool):
        self.emitter = emitter
        self.repeating = repeating
        self.timer: Optional[threading.Timer] = None
        self.lock = threading.RLock()

    def start(self, timeout: float) -> bool:
        with self.lock:
            if self.timer is not None:
                return False
            else:
                self._schedule_timer(timeout=timeout)
                return True

    def cancel(self) -> None:
        with self.lock:
            if self.timer is not None:
                self.timer.cancel()
                self.timer = None

    def is_active(self) -> bool:
        with self.lock:
            return self.timer is not None

    def _fire(self, timeout: float) -> None:
        with self.lock:
            if self.repeating:
                self._schedule_timer(timeout)
            else:
                self.timer = None

        self.emitter.flush()

    def _schedule_timer(self, timeout: float) -> None:
        self.timer = threading.Timer(timeout, self._fire, [timeout])
        self.timer.daemon = True
        self.timer.start()



# """
#     tracker.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import time
import uuid
from typing import Any, Optional, Union, List, Dict, Sequence
from warnings import warn

from snowplow_tracker import payload, SelfDescribingJson
from snowplow_tracker.subject import Subject
from snowplow_tracker.contracts import non_empty_string, one_of, non_empty, form_element
from snowplow_tracker.constants import (
    VERSION,
    DEFAULT_ENCODE_BASE64,
    BASE_SCHEMA_PATH,
    SCHEMA_TAG,
)

from snowplow_tracker.events import (
    Event,
    PagePing,
    PageView,
    SelfDescribing,
    StructuredEvent,
    ScreenView,
)
from snowplow_tracker.typing import (
    JsonEncoderFunction,
    EmitterProtocol,
    FORM_NODE_NAMES,
    FORM_TYPES,
    FormNodeName,
    ElementClasses,
    FormClasses,
)

"""
Tracker class
"""


class Tracker:
    def __init__(
        self,
        namespace: str,
        emitters: Union[List[EmitterProtocol], EmitterProtocol],
        subject: Optional[Subject] = None,
        app_id: Optional[str] = None,
        encode_base64: bool = DEFAULT_ENCODE_BASE64,
        json_encoder: Optional[JsonEncoderFunction] = None,
    ) -> None:
        """
        :param namespace:        Identifier for the Tracker instance
        :type  namespace:        string
        :param emitters:         Emitters to which events will be sent
        :type  emitters:         list[>0](emitter) | emitter
        :param subject:          Subject to be tracked
        :type  subject:          subject | None
        :param app_id:           Application ID
        :type  app_id:           string_or_none
        :param encode_base64:    Whether JSONs in the payload should be base-64 encoded
        :type  encode_base64:    bool
        :param json_encoder:     Custom JSON serializer that gets called on non-serializable object
        :type  json_encoder:     function | None
        """
        if subject is None:
            subject = Subject()

        if isinstance(emitters, list):
            non_empty(emitters)
            self.emitters = emitters
        else:
            self.emitters = [emitters]

        self.subject: Optional[Subject] = subject
        self.encode_base64 = encode_base64
        self.json_encoder = json_encoder

        self.standard_nv_pairs = {"tv": VERSION, "tna": namespace, "aid": app_id}
        self.timer = None

    @staticmethod
    def get_uuid() -> str:
        """
        Set transaction ID for the payload once during the lifetime of the
        event.

        :rtype:           string
        """
        return str(uuid.uuid4())

    @staticmethod
    def get_timestamp(tstamp: Optional[float] = None) -> int:
        """
        :param tstamp:    User-input timestamp or None
        :type  tstamp:    int | float | None
        :rtype:           int
        """
        if isinstance(
            tstamp,
            (
                int,
                float,
            ),
        ):
            return int(tstamp)
        return int(time.time() * 1000)

    """
    Tracking methods
    """

    def track(
        self,
        event: Event,
    ) -> Optional[str]:
        """
        Send the event payload to a emitter. Returns the tracked event ID.
        :param  event:           Event
        :type   event:           events.Event
        :rtype:                  String
        """

        payload = self.complete_payload(
            event=event,
        )

        for emitter in self.emitters:
            emitter.input(payload.nv_pairs)

        if "eid" in payload.nv_pairs.keys():
            return payload.nv_pairs["eid"]

        return None

    def complete_payload(
        self,
        event: Event,
    ) -> payload.Payload:
        payload = event.build_payload(
            encode_base64=self.encode_base64,
            json_encoder=self.json_encoder,
            subject=self.subject,
        )

        payload.add("eid", Tracker.get_uuid())
        payload.add("dtm", Tracker.get_timestamp())
        payload.add_dict(self.standard_nv_pairs)

        return payload

    def track_page_view(
        self,
        page_url: str,
        page_title: Optional[str] = None,
        referrer: Optional[str] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  page_url:       URL of the viewed page
        :type   page_url:       non_empty_string
        :param  page_title:     Title of the viewed page
        :type   page_title:     string_or_none
        :param  referrer:       Referrer of the page
        :type   referrer:       string_or_none
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_page_view will be removed in future versions. Please use the new PageView class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )

        pv = PageView(
            page_url=page_url,
            page_title=page_title,
            referrer=referrer,
            event_subject=event_subject,
            context=context,
            true_timestamp=tstamp,
        )

        self.track(event=pv)
        return self

    def track_page_ping(
        self,
        page_url: str,
        page_title: Optional[str] = None,
        referrer: Optional[str] = None,
        min_x: Optional[int] = None,
        max_x: Optional[int] = None,
        min_y: Optional[int] = None,
        max_y: Optional[int] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  page_url:       URL of the viewed page
        :type   page_url:       non_empty_string
        :param  page_title:     Title of the viewed page
        :type   page_title:     string_or_none
        :param  referrer:       Referrer of the page
        :type   referrer:       string_or_none
        :param  min_x:          Minimum page x offset seen in the last ping period
        :type   min_x:          int | None
        :param  max_x:          Maximum page x offset seen in the last ping period
        :type   max_x:          int | None
        :param  min_y:          Minimum page y offset seen in the last ping period
        :type   min_y:          int | None
        :param  max_y:          Maximum page y offset seen in the last ping period
        :type   max_y:          int | None
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_page_ping will be removed in future versions. Please use the new PagePing class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )

        pp = PagePing(
            page_url=page_url,
            page_title=page_title,
            referrer=referrer,
            min_x=min_x,
            max_x=max_x,
            min_y=min_y,
            max_y=max_y,
            context=context,
            true_timestamp=tstamp,
            event_subject=event_subject,
        )

        self.track(event=pp)
        return self

    def track_link_click(
        self,
        target_url: str,
        element_id: Optional[str] = None,
        element_classes: Optional[ElementClasses] = None,
        element_target: Optional[str] = None,
        element_content: Optional[str] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  target_url:         Target URL of the link
        :type   target_url:         non_empty_string
        :param  element_id:         ID attribute of the HTML element
        :type   element_id:         string_or_none
        :param  element_classes:    Classes of the HTML element
        :type   element_classes:    list(str) | tuple(str,\\*) | None
        :param  element_target:     ID attribute of the HTML element
        :type   element_target:     string_or_none
        :param  element_content:    The content of the HTML element
        :type   element_content:    string_or_none
        :param  context:            Custom context for the event
        :type   context:            context_array | None
        :param  tstamp:             Optional event timestamp in milliseconds
        :type   tstamp:             int | float | None
        :param  event_subject:      Optional per event subject
        :type   event_subject:      subject | None
        :rtype:                     Tracker
        """
        warn(
            "track_link_click will be removed in future versions. Please use the new SelfDescribing class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )
        non_empty_string(target_url)

        properties: Dict[str, Union[str, ElementClasses]] = {}
        properties["targetUrl"] = target_url
        if element_id is not None:
            properties["elementId"] = element_id
        if element_classes is not None:
            properties["elementClasses"] = element_classes
        if element_target is not None:
            properties["elementTarget"] = element_target
        if element_content is not None:
            properties["elementContent"] = element_content

        event_json = SelfDescribingJson(
            "%s/link_click/%s/1-0-1" % (BASE_SCHEMA_PATH, SCHEMA_TAG), properties
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def track_add_to_cart(
        self,
        sku: str,
        quantity: int,
        name: Optional[str] = None,
        category: Optional[str] = None,
        unit_price: Optional[float] = None,
        currency: Optional[str] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  sku:            Item SKU or ID
        :type   sku:            non_empty_string
        :param  quantity:       Number added to cart
        :type   quantity:       int
        :param  name:           Item's name
        :type   name:           string_or_none
        :param  category:       Item's category
        :type   category:       string_or_none
        :param  unit_price:     Item's price
        :type   unit_price:     int | float | None
        :param  currency:       Type of currency the price is in
        :type   currency:       string_or_none
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_add_to_cart will be deprecated in future versions.",
            DeprecationWarning,
            stacklevel=2,
        )
        non_empty_string(sku)

        properties: Union[Dict[str, Union[str, float, int]]] = {}
        properties["sku"] = sku
        properties["quantity"] = quantity
        if name is not None:
            properties["name"] = name
        if category is not None:
            properties["category"] = category
        if unit_price is not None:
            properties["unitPrice"] = unit_price
        if currency is not None:
            properties["currency"] = currency

        event_json = SelfDescribingJson(
            "%s/add_to_cart/%s/1-0-0" % (BASE_SCHEMA_PATH, SCHEMA_TAG), properties
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def track_remove_from_cart(
        self,
        sku: str,
        quantity: int,
        name: Optional[str] = None,
        category: Optional[str] = None,
        unit_price: Optional[float] = None,
        currency: Optional[str] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  sku:            Item SKU or ID
        :type   sku:            non_empty_string
        :param  quantity:       Number added to cart
        :type   quantity:       int
        :param  name:           Item's name
        :type   name:           string_or_none
        :param  category:       Item's category
        :type   category:       string_or_none
        :param  unit_price:     Item's price
        :type   unit_price:     int | float | None
        :param  currency:       Type of currency the price is in
        :type   currency:       string_or_none
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_remove_from_cart will be deprecated in future versions.",
            DeprecationWarning,
            stacklevel=2,
        )
        non_empty_string(sku)

        properties: Dict[str, Union[str, float, int]] = {}
        properties["sku"] = sku
        properties["quantity"] = quantity
        if name is not None:
            properties["name"] = name
        if category is not None:
            properties["category"] = category
        if unit_price is not None:
            properties["unitPrice"] = unit_price
        if currency is not None:
            properties["currency"] = currency

        event_json = SelfDescribingJson(
            "%s/remove_from_cart/%s/1-0-0" % (BASE_SCHEMA_PATH, SCHEMA_TAG), properties
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def track_form_change(
        self,
        form_id: str,
        element_id: Optional[str],
        node_name: FormNodeName,
        value: Optional[str],
        type_: Optional[str] = None,
        element_classes: Optional[ElementClasses] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  form_id:            ID attribute of the HTML form
        :type   form_id:            non_empty_string
        :param  element_id:         ID attribute of the HTML element
        :type   element_id:         string_or_none
        :param  node_name:          Type of input element
        :type   node_name:          form_node_name
        :param  value:              Value of the input element
        :type   value:              string_or_none
        :param  type_:              Type of data the element represents
        :type   type_:              non_empty_string, form_type
        :param  element_classes:    Classes of the HTML element
        :type   element_classes:    list(str) | tuple(str,\\*) | None
        :param  context:            Custom context for the event
        :type   context:            context_array | None
        :param  tstamp:             Optional event timestamp in milliseconds
        :type   tstamp:             int | float | None
        :param  event_subject:      Optional per event subject
        :type   event_subject:      subject | None
        :rtype:                     Tracker
        """
        warn(
            "track_form_change will be removed in future versions. Please use the new SelfDescribing class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )

        non_empty_string(form_id)
        one_of(node_name, FORM_NODE_NAMES)
        if type_ is not None:
            one_of(type_.lower(), FORM_TYPES)

        properties: Dict[str, Union[Optional[str], ElementClasses]] = dict()
        properties["formId"] = form_id
        properties["elementId"] = element_id
        properties["nodeName"] = node_name
        properties["value"] = value
        if type_ is not None:
            properties["type"] = type_
        if element_classes is not None:
            properties["elementClasses"] = element_classes

        event_json = SelfDescribingJson(
            "%s/change_form/%s/1-0-0" % (BASE_SCHEMA_PATH, SCHEMA_TAG), properties
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def track_form_submit(
        self,
        form_id: str,
        form_classes: Optional[FormClasses] = None,
        elements: Optional[List[Dict[str, Any]]] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  form_id:        ID attribute of the HTML form
        :type   form_id:        non_empty_string
        :param  form_classes:   Classes of the HTML form
        :type   form_classes:   list(str) | tuple(str,\\*) | None
        :param  elements:       Classes of the HTML form
        :type   elements:       list(form_element) | None
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_form_submit will be removed in future versions. Please use the new SelfDescribing class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )
        non_empty_string(form_id)

        for element in elements or []:
            form_element(element)

        properties: Dict[
            str, Union[str, ElementClasses, FormClasses, List[Dict[str, Any]]]
        ] = dict()
        properties["formId"] = form_id
        if form_classes is not None:
            properties["formClasses"] = form_classes
        if elements is not None and len(elements) > 0:
            properties["elements"] = elements

        event_json = SelfDescribingJson(
            "%s/submit_form/%s/1-0-0" % (BASE_SCHEMA_PATH, SCHEMA_TAG), properties
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def track_site_search(
        self,
        terms: Sequence[str],
        filters: Optional[Dict[str, Union[str, bool]]] = None,
        total_results: Optional[int] = None,
        page_results: Optional[int] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  terms:          Search terms
        :type   terms:          seq[>=1](str)
        :param  filters:        Filters applied to the search
        :type   filters:        dict(str:str|bool) | None
        :param  total_results:  Total number of results returned
        :type   total_results:  int | None
        :param  page_results:   Total number of pages of results
        :type   page_results:   int | None
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_site_search will be removed in future versions. Please use the new SelfDescribing class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )
        non_empty(terms)

        properties: Dict[
            str, Union[Sequence[str], Dict[str, Union[str, bool]], int]
        ] = {}
        properties["terms"] = terms
        if filters is not None:
            properties["filters"] = filters
        if total_results is not None:
            properties["totalResults"] = total_results
        if page_results is not None:
            properties["pageResults"] = page_results

        event_json = SelfDescribingJson(
            "%s/site_search/%s/1-0-0" % (BASE_SCHEMA_PATH, SCHEMA_TAG), properties
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def track_ecommerce_transaction_item(
        self,
        order_id: str,
        sku: str,
        price: float,
        quantity: int,
        name: Optional[str] = None,
        category: Optional[str] = None,
        currency: Optional[str] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        This is an internal method called by track_ecommerce_transaction.
        It is not for public use.

        :param  order_id:       Order ID
        :type   order_id:       non_empty_string
        :param  sku:            Item SKU
        :type   sku:            non_empty_string
        :param  price:          Item price
        :type   price:          int | float
        :param  quantity:       Item quantity
        :type   quantity:       int
        :param  name:           Item name
        :type   name:           string_or_none
        :param  category:       Item category
        :type   category:       string_or_none
        :param  currency:       The currency the price is expressed in
        :type   currency:       string_or_none
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_ecommerce_transaction_item will be deprecated in future versions.",
            DeprecationWarning,
            stacklevel=2,
        )
        non_empty_string(order_id)
        non_empty_string(sku)

        event = Event(
            event_subject=event_subject, context=context, true_timestamp=tstamp
        )
        event.payload.add("e", "ti")
        event.payload.add("ti_id", order_id)
        event.payload.add("ti_sk", sku)
        event.payload.add("ti_nm", name)
        event.payload.add("ti_ca", category)
        event.payload.add("ti_pr", price)
        event.payload.add("ti_qu", quantity)
        event.payload.add("ti_cu", currency)

        self.track(event=event)
        return self

    def track_ecommerce_transaction(
        self,
        order_id: str,
        total_value: float,
        affiliation: Optional[str] = None,
        tax_value: Optional[float] = None,
        shipping: Optional[float] = None,
        city: Optional[str] = None,
        state: Optional[str] = None,
        country: Optional[str] = None,
        currency: Optional[str] = None,
        items: Optional[List[Dict[str, Any]]] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  order_id:       ID of the eCommerce transaction
        :type   order_id:       non_empty_string
        :param  total_value:    Total transaction value
        :type   total_value:    int | float
        :param  affiliation:    Transaction affiliation
        :type   affiliation:    string_or_none
        :param  tax_value:      Transaction tax value
        :type   tax_value:      int | float | None
        :param  shipping:       Delivery cost charged
        :type   shipping:       int | float | None
        :param  city:           Delivery address city
        :type   city:           string_or_none
        :param  state:          Delivery address state
        :type   state:          string_or_none
        :param  country:        Delivery address country
        :type   country:        string_or_none
        :param  currency:       The currency the price is expressed in
        :type   currency:       string_or_none
        :param  items:          The items in the transaction
        :type   items:          list(dict(str:\\*)) | None
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_ecommerce_transaction will be deprecated in future versions.",
            DeprecationWarning,
            stacklevel=2,
        )
        non_empty_string(order_id)

        event = Event(
            event_subject=event_subject, context=context, true_timestamp=tstamp
        )
        event.payload.add("e", "tr")
        event.payload.add("tr_id", order_id)
        event.payload.add("tr_tt", total_value)
        event.payload.add("tr_af", affiliation)
        event.payload.add("tr_tx", tax_value)
        event.payload.add("tr_sh", shipping)
        event.payload.add("tr_ci", city)
        event.payload.add("tr_st", state)
        event.payload.add("tr_co", country)
        event.payload.add("tr_cu", currency)

        tstamp = Tracker.get_timestamp(tstamp)

        self.track(event=event)

        if items is None:
            items = []
        for item in items:
            item["order_id"] = order_id
            item["currency"] = currency
            item["tstamp"] = tstamp
            item["event_subject"] = event_subject
            item["context"] = context
            self.track_ecommerce_transaction_item(**item)

        return self

    def track_screen_view(
        self,
        name: Optional[str] = None,
        id_: Optional[str] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  name:           The name of the screen view event
        :type   name:           string_or_none
        :param  id_:            Screen view ID
        :type   id_:            string_or_none
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_screen_view will be removed in future versions. Please use the new ScreenView class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )
        screen_view_properties = {}
        if name is not None:
            screen_view_properties["name"] = name
        if id_ is not None:
            screen_view_properties["id"] = id_

        event_json = SelfDescribingJson(
            "%s/screen_view/%s/1-0-0" % (BASE_SCHEMA_PATH, SCHEMA_TAG),
            screen_view_properties,
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def track_mobile_screen_view(
        self,
        name: str,
        id_: Optional[str] = None,
        type: Optional[str] = None,
        previous_name: Optional[str] = None,
        previous_id: Optional[str] = None,
        previous_type: Optional[str] = None,
        transition_type: Optional[str] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  name:           The name of the screen view event
        :type   name:           string_or_none
        :param  id_:            Screen view ID. This must be of type UUID.
        :type   id_:            string | None
        :param  type:           The type of screen that was viewed e.g feed / carousel.
        :type   type:           string | None
        :param  previous_name:  The name of the previous screen.
        :type   previous_name:  string | None
        :param  previous_id:    The screenview ID of the previous screenview.
        :type   previous_id:    string | None
        :param  previous_type   The screen type of the previous screenview
        :type   previous_type   string | None
        :param  transition_type The type of transition that led to the screen being viewed.
        :type   transition_type string | None
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_mobile_screen_view will be removed in future versions. Please use the new ScreenView class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )
        if id_ is None:
            id_ = self.get_uuid()

        sv = ScreenView(
            name=name,
            id_=id_,
            type=type,
            previous_name=previous_name,
            previous_id=previous_id,
            previous_type=previous_type,
            transition_type=transition_type,
            event_subject=event_subject,
            context=context,
            true_timestamp=tstamp,
        )

        self.track(event=sv)
        return self

    def track_struct_event(
        self,
        category: str,
        action: str,
        label: Optional[str] = None,
        property_: Optional[str] = None,
        value: Optional[Union[int, float]] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  category:       Category of the event
        :type   category:       non_empty_string
        :param  action:         The event itself
        :type   action:         non_empty_string
        :param  label:          Refer to the object the action is
                                performed on
        :type   label:          string_or_none
        :param  property_:      Property associated with either the action
                                or the object
        :type   property_:      string_or_none
        :param  value:          A value associated with the user action
        :type   value:          int | float | None
        :param  context:        Custom context for the event
        :type   context:        context_array | None
        :param  tstamp:         Optional event timestamp in milliseconds
        :type   tstamp:         int | float | None
        :param  event_subject:  Optional per event subject
        :type   event_subject:  subject | None
        :rtype:                 Tracker
        """
        warn(
            "track_struct_event will be removed in future versions. Please use the new Structured class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )
        se = StructuredEvent(
            category=category,
            action=action,
            label=label,
            property_=property_,
            value=value,
            context=context,
            true_timestamp=tstamp,
            event_subject=event_subject,
        )

        self.track(
            event=se,
        )
        return self

    def track_self_describing_event(
        self,
        event_json: SelfDescribingJson,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  event_json:      The properties of the event. Has two field:
                                 A "data" field containing the event properties and
                                 A "schema" field identifying the schema against which the data is validated
        :type   event_json:      self_describing_json
        :param  context:         Custom context for the event
        :type   context:         context_array | None
        :param  tstamp:          Optional event timestamp in milliseconds
        :type   tstamp:          int | float | None
        :param  event_subject:   Optional per event subject
        :type   event_subject:   subject | None
        :rtype:                  Tracker
        """
        warn(
            "track_self_describing_event will be removed in future versions. Please use the new SelfDescribing class to track the event.",
            DeprecationWarning,
            stacklevel=2,
        )

        sd = SelfDescribing(
            event_json=event_json,
            context=context,
            true_timestamp=tstamp,
            event_subject=event_subject,
        )
        self.track(
            event=sd,
        )
        return self

    # Alias
    def track_unstruct_event(
        self,
        event_json: SelfDescribingJson,
        context: Optional[List[SelfDescribingJson]] = None,
        tstamp: Optional[float] = None,
        event_subject: Optional[Subject] = None,
    ) -> "Tracker":
        """
        :param  event_json:      The properties of the event. Has two field:
                                 A "data" field containing the event properties and
                                 A "schema" field identifying the schema against which the data is validated
        :type   event_json:      self_describing_json
        :param  context:         Custom context for the event
        :type   context:         context_array | None
        :param  tstamp:          Optional event timestamp in milliseconds
        :type   tstamp:          int | float | None
        :param  event_subject:   Optional per event subject
        :type   event_subject:   subject | None
        :rtype:                  Tracker
        """
        warn(
            "track_unstruct_event will be deprecated in future versions. Please use track_self_describing_event.",
            DeprecationWarning,
            stacklevel=2,
        )

        self.track_self_describing_event(
            event_json=event_json,
            context=context,
            tstamp=tstamp,
            event_subject=event_subject,
        )
        return self

    def flush(self, is_async: bool = False) -> "Tracker":
        """
        Flush the emitter

        :param  is_async:  Whether the flush is done asynchronously. Default is False
        :type   is_async:  bool
        :rtype:         tracker
        """
        for emitter in self.emitters:
            if is_async:
                if hasattr(emitter, "flush"):
                    emitter.flush()
            else:
                if hasattr(emitter, "sync_flush"):
                    emitter.sync_flush()
        return self

    def set_subject(self, subject: Optional[Subject]) -> "Tracker":
        """
        Set the subject of the events fired by the tracker

        :param subject: Subject to be tracked
        :type  subject: subject | None
        :rtype:         tracker
        """
        self.subject = subject
        return self

    def add_emitter(self, emitter: EmitterProtocol) -> "Tracker":
        """
        Add a new emitter to which events should be passed

        :param emitter: New emitter
        :type  emitter: emitter
        :rtype:         tracker
        """
        self.emitters.append(emitter)
        return self

    def get_namespace(self) -> str:
        # As app_id is added to the standard_nv_pairs dict above with a type of Optional[str], the type for
        # the whole standard_nv_pairs dict is inferred to be dict[str, Optional[str]].
        # But, we know that "tna" should always be present in the dict, since namespace is a required argument.
        #
        # This ignores MyPy saying Incompatible return value type (got "str | None", expected "str")
        return self.standard_nv_pairs["tna"]  # type: ignore



# """
#     subject.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

from typing import Dict, Optional, Union
from snowplow_tracker.contracts import one_of, greater_than
from snowplow_tracker.typing import SupportedPlatform, SUPPORTED_PLATFORMS, PayloadDict

DEFAULT_PLATFORM = "pc"


class Subject(object):
    """
    Class for an event subject, where we view events as of the form

    (Subject) -> (Verb) -> (Object)
    """

    def __init__(self) -> None:
        self.standard_nv_pairs: Dict[str, Union[str, int]] = {"p": DEFAULT_PLATFORM}

    def set_platform(self, value: SupportedPlatform) -> "Subject":
        """
        :param  value:          One of ["pc", "tv", "mob", "cnsl", "iot", "web", "srv", "app"]
        :type   value:          supported_platform
        :rtype:                 subject
        """
        one_of(value, SUPPORTED_PLATFORMS)

        self.standard_nv_pairs["p"] = value
        return self

    def set_user_id(self, user_id: str) -> "Subject":
        """
        :param  user_id:        User ID
        :type   user_id:        string
        :rtype:                 subject
        """
        self.standard_nv_pairs["uid"] = user_id
        return self

    def set_screen_resolution(self, width: int, height: int) -> "Subject":
        """
        :param  width:          Width of the screen
        :param  height:         Height of the screen
        :type   width:          int,>0
        :type   height:         int,>0
        :rtype:                 subject
        """
        greater_than(width, 0)
        greater_than(height, 0)

        self.standard_nv_pairs["res"] = "".join([str(width), "x", str(height)])
        return self

    def set_viewport(self, width: int, height: int) -> "Subject":
        """
        :param  width:          Width of the viewport
        :param  height:         Height of the viewport
        :type   width:          int,>0
        :type   height:         int,>0
        :rtype:                 subject
        """
        greater_than(width, 0)
        greater_than(height, 0)

        self.standard_nv_pairs["vp"] = "".join([str(width), "x", str(height)])
        return self

    def set_color_depth(self, depth: int) -> "Subject":
        """
        :param  depth:          Depth of the color on the screen
        :type   depth:          int
        :rtype:                 subject
        """
        self.standard_nv_pairs["cd"] = depth
        return self

    def set_timezone(self, timezone: str) -> "Subject":
        """
        :param  timezone:       Timezone as a string
        :type   timezone:       string
        :rtype:                 subject
        """
        self.standard_nv_pairs["tz"] = timezone
        return self

    def set_lang(self, lang: str) -> "Subject":
        """
        Set language.

        :param  lang:           Language the application is set to
        :type   lang:           string
        :rtype:                 subject
        """
        self.standard_nv_pairs["lang"] = lang
        return self

    def set_domain_user_id(self, duid: str) -> "Subject":
        """
        Set the domain user ID

        :param duid:            Domain user ID
        :type  duid:            string
        :rtype:                 subject
        """
        self.standard_nv_pairs["duid"] = duid
        return self

    def set_domain_session_id(self, sid: str) -> "Subject":
        """
        Set the domain session ID
        :param sid:             Domain session ID
        :type  sid:             string
        :rtype:                 subject
        """
        self.standard_nv_pairs["sid"] = sid
        return self

    def set_domain_session_index(self, vid: int) -> "Subject":
        """
        Set the domain session Index
        :param vid:             Domain session Index
        :type vid:              int
        :rtype:                 subject
        """
        self.standard_nv_pairs["vid"] = vid
        return self

    def set_ip_address(self, ip: str) -> "Subject":
        """
        Set the domain user ID

        :param ip:              IP address
        :type  ip:              string
        :rtype:                 subject
        """
        self.standard_nv_pairs["ip"] = ip
        return self

    def set_useragent(self, ua: str) -> "Subject":
        """
        Set the user agent

        :param ua:              User agent
        :type  ua:              string
        :rtype:                 subject
        """
        self.standard_nv_pairs["ua"] = ua
        return self

    def set_network_user_id(self, nuid: str) -> "Subject":
        """
        Set the network user ID field
        This overwrites the nuid field set by the collector

        :param nuid:            Network user ID
        :type  nuid:            string
        :rtype:                 subject
        """
        self.standard_nv_pairs["tnuid"] = nuid
        return self

    def combine_subject(self, subject: Optional["Subject"]) -> PayloadDict:
        """
        Merges another instance of Subject, with self taking priority
        :param  subject     Subject to update
        :type   subject     subject
        :rtype              PayloadDict

        """
        if subject is not None:
            return {**subject.standard_nv_pairs, **self.standard_nv_pairs}

        return self.standard_nv_pairs



# """
#     emitter_configuration.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

from typing import Optional, Union, Tuple, Dict
from snowplow_tracker.typing import SuccessCallback, FailureCallback
from snowplow_tracker.event_store import EventStore
import requests


class EmitterConfiguration(object):
    def __init__(
        self,
        batch_size: Optional[int] = None,
        on_success: Optional[SuccessCallback] = None,
        on_failure: Optional[FailureCallback] = None,
        byte_limit: Optional[int] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        buffer_capacity: Optional[int] = None,
        custom_retry_codes: Dict[int, bool] = {},
        event_store: Optional[EventStore] = None,
        session: Optional[requests.Session] = None,
    ) -> None:
        """
        Configuration for the emitter that sends events to the Snowplow collector.
        :param batch_size:     The maximum number of queued events before the buffer is flushed. Default is 10.
        :type  batch_size:     int | None
        :param on_success:      Callback executed after every HTTP request in a flush has status code 200
                                Gets passed one argument, an array of dictionaries corresponding to the sent events' payloads
        :type  on_success:      function | None
        :param on_failure:      Callback executed if at least one HTTP request in a flush has status code other than 200
                                Gets passed two arguments:
                                1) The number of events which were successfully sent
                                2) An array of dictionaries corresponding to the unsent events' payloads
        :type  on_failure:      function | None
        :param byte_limit:      The size event list after reaching which queued events will be flushed
        :type  byte_limit:      int | None
        :param request_timeout: Timeout for the HTTP requests. Can be set either as single float value which
                                 applies to both "connect" AND "read" timeout, or as tuple with two float values
                                 which specify the "connect" and "read" timeouts separately
        :type request_timeout:  float | tuple | None
        :param  custom_retry_codes: Set custom retry rules for HTTP status codes received in emit responses from the Collector.
                                    By default, retry will not occur for status codes 400, 401, 403, 410 or 422. This can be overridden here.
                                    Note that 2xx codes will never retry as they are considered successful.
        :type   custom_retry_codes: dict
        :param  event_store:    Stores the event buffer and buffer capacity. Default is an InMemoryEventStore object with buffer_capacity of 10,000 events.
        :type   event_store:    EventStore | None
        :param  session:    Persist parameters across requests by using a session object
        :type   session:    request.Session | None
        """

        self.batch_size = batch_size
        self.on_success = on_success
        self.on_failure = on_failure
        self.byte_limit = byte_limit
        self.request_timeout = request_timeout
        self.buffer_capacity = buffer_capacity
        self.custom_retry_codes = custom_retry_codes
        self.event_store = event_store
        self.session = session

    @property
    def batch_size(self) -> Optional[int]:
        """
        The maximum number of queued events before the buffer is flushed. Default is 10.
        """
        return self._batch_size

    @batch_size.setter
    def batch_size(self, value: Optional[int]):
        if isinstance(value, int) and value < 0:
            raise ValueError("batch_size must greater than 0")
        if not isinstance(value, int) and value is not None:
            raise ValueError("batch_size must be of type int")
        self._batch_size = value

    @property
    def on_success(self) -> Optional[SuccessCallback]:
        """
        Callback executed after every HTTP request in a flush has status code 200. Gets passed the number of events flushed.
        """
        return self._on_success

    @on_success.setter
    def on_success(self, value: Optional[SuccessCallback]):
        self._on_success = value

    @property
    def on_failure(self) -> Optional[FailureCallback]:
        """
        Callback executed if at least one HTTP request in a flush has status code other than 200
                                Gets passed two arguments:
                                1) The number of events which were successfully sent
                                2) An array of dictionaries corresponding to the unsent events' payloads
        """
        return self._on_failure

    @on_failure.setter
    def on_failure(self, value: Optional[FailureCallback]):
        self._on_failure = value

    @property
    def byte_limit(self) -> Optional[int]:
        """
        The size event list after reaching which queued events will be flushed
        """
        return self._byte_limit

    @byte_limit.setter
    def byte_limit(self, value: Optional[int]):
        if isinstance(value, int) and value < 0:
            raise ValueError("byte_limit must greater than 0")
        if not isinstance(value, int) and value is not None:
            raise ValueError("byte_limit must be of type int")
        self._byte_limit = value

    @property
    def request_timeout(self) -> Optional[Union[float, Tuple[float, float]]]:
        """
        Timeout for the HTTP requests. Can be set either as single float value which
                                     applies to both "connect" AND "read" timeout, or as tuple with two float values
                                     which specify the "connect" and "read" timeouts separately
        """
        return self._request_timeout

    @request_timeout.setter
    def request_timeout(self, value: Optional[Union[float, Tuple[float, float]]]):
        self._request_timeout = value

    @property
    def buffer_capacity(self) -> Optional[int]:
        """
        The maximum capacity of the event buffer. The default buffer capacity is 10 000 events.
                                When the buffer is full new events are lost.
        """
        return self._buffer_capacity

    @buffer_capacity.setter
    def buffer_capacity(self, value: Optional[int]):
        if isinstance(value, int) and value < 0:
            raise ValueError("buffer_capacity must greater than 0")
        if not isinstance(value, int) and value is not None:
            raise ValueError("buffer_capacity must be of type int")
        self._buffer_capacity = value

    @property
    def custom_retry_codes(self) -> Dict[int, bool]:
        """
        Custom retry rules for HTTP status codes received in emit responses from the Collector.
        """
        return self._custom_retry_codes

    @custom_retry_codes.setter
    def custom_retry_codes(self, value: Dict[int, bool]):
        self._custom_retry_codes = value

    def set_retry_code(self, status_code: int, retry=True) -> bool:
        """
        Add a retry rule for HTTP status code received from emit responses from the Collector.
        :param  status_code:    HTTP response code
        :type   status_code:    int
        :param  retry:  Set the status_code to retry (True) or not retry (False). Default is True
        :type   retry:  bool
        """
        if not isinstance(status_code, int):
            print("status_code must be of type int")
            return False

        if not isinstance(retry, bool):
            print("retry must be of type bool")
            return False

        if 200 <= status_code < 300:
            print(
                "custom_retry_codes should not include codes for succesful requests (2XX codes)"
            )
            return False

        self.custom_retry_codes[status_code] = retry

        return status_code in self.custom_retry_codes.keys()

    @property
    def event_store(self) -> Optional[EventStore]:
        return self._event_store

    @event_store.setter
    def event_store(self, value: Optional[EventStore]):
        self._event_store = value

    @property
    def session(self) -> Optional[requests.Session]:
        """
        Persist parameters across requests using a requests.Session object
        """
        return self._session

    @session.setter
    def session(self, value: Optional[requests.Session]):
        self._session = value



# """
#     snowplow.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import logging
from typing import Dict, Optional
from snowplow_tracker import (
    Tracker,
    Emitter,
    subject,
    EmitterConfiguration,
    TrackerConfiguration,
)
from snowplow_tracker.typing import Method

# Logging
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

"""
Snowplow Class
"""


class Snowplow:
    _trackers: Dict[str, Tracker] = {}

    @staticmethod
    def create_tracker(
        namespace: str,
        endpoint: str,
        method: Method = "post",
        app_id: Optional[str] = None,
        subject: Optional[subject.Subject] = None,
        tracker_config: TrackerConfiguration = TrackerConfiguration(),
        emitter_config: EmitterConfiguration = EmitterConfiguration(),
    ) -> Tracker:
        """
        Create a Snowplow tracker with a namespace and collector URL

        :param  namespace:          Name of the tracker
        :type   namespace:          String
        :param  endpoint:           The collector URL
        :type   endpoint:           String
        :param  method:             The HTTP request method. Defaults to post.
        :type   method:             method
        :param  appId:              Application ID
        :type   appId:              String | None
        :param  subject:            Subject to be tracked
        :type   subject:            Subject | None
        :param  tracker_config:     Tracker configuration
        :type   tracker_config:     TrackerConfiguration
        :param  emitter_config:     Emitter configuration
        :type   emitter_config:     EmitterConfiguration
        :rtype                      Tracker
        """
        if endpoint is None:
            raise TypeError("Emitter or Collector URL must be provided")

        emitter = Emitter(
            endpoint=endpoint,
            method=method,
            batch_size=emitter_config.batch_size,
            on_success=emitter_config.on_success,
            on_failure=emitter_config.on_failure,
            byte_limit=emitter_config.byte_limit,
            request_timeout=emitter_config.request_timeout,
            custom_retry_codes=emitter_config.custom_retry_codes,
            event_store=emitter_config.event_store,
            session=emitter_config.session,
        )

        tracker = Tracker(
            namespace=namespace,
            emitters=emitter,
            app_id=app_id,
            subject=subject,
            encode_base64=tracker_config.encode_base64,
            json_encoder=tracker_config.json_encoder,
        )

        return Snowplow.add_tracker(tracker)

    @classmethod
    def add_tracker(cls, tracker: Tracker) -> Tracker:
        """
        Add a Snowplow tracker to the Snowplow object

        :param  tracker:  Tracker object to add to Snowplow
        :type   tracker:  Tracker
        :rtype            Tracker
        """
        if not isinstance(tracker, Tracker):
            logger.info("Tracker not provided.")
            return None

        namespace = tracker.get_namespace()

        if namespace in cls._trackers.keys():
            raise TypeError("Tracker with this namespace already exists")

        cls._trackers[namespace] = tracker
        logger.info("Tracker with namespace: '" + namespace + "' added to Snowplow")
        return cls._trackers[namespace]

    @classmethod
    def remove_tracker(cls, tracker: Tracker):
        """
        Remove a Snowplow tracker from the Snowplow object if it exists

        :param  tracker:        Tracker object to remove from Snowplow
        :type   tracker:        Tracker | None
        """
        namespace = tracker.get_namespace()
        cls.remove_tracker_by_namespace(namespace)

    @classmethod
    def remove_tracker_by_namespace(cls, namespace: str):
        """
        Remove a Snowplow tracker from the Snowplow object using it's namespace if it exists

        :param  namespace:      Tracker namespace to remove from Snowplow
        :type   tracker:        String | None
        """
        if not cls._trackers.pop(namespace, False):
            logger.info("Tracker with namespace: '" + namespace + "' does not exist")
            return
        logger.info("Tracker with namespace: '" + namespace + "' removed from Snowplow")

    @classmethod
    def reset(cls):
        """
        Remove all active Snowplow trackers from the Snowplow object
        """
        cls._trackers = {}

    @classmethod
    def get_tracker(cls, namespace: str) -> Optional[Tracker]:
        """
        Returns a Snowplow tracker from the Snowplow object if it exists
        :param  namespace:              Snowplow tracker namespace
        :type   namespace:              string
        :rtype:                         Tracker
        """
        if namespace in cls._trackers.keys():
            return cls._trackers[namespace]
        return None



# """
#     event_store.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

from typing import List
from typing_extensions import Protocol
from snowplow_tracker.typing import PayloadDict, PayloadDictList
from logging import Logger


class EventStore(Protocol):
    """
    EventStore protocol. For buffering events in the Emitter.
    """

    def add_event(self, payload: PayloadDict) -> bool:
        """
        Add PayloadDict to buffer. Returns True if successful.

        :param payload: The payload to add
        :type  payload: PayloadDict
        :rtype  bool
        """
        ...

    def get_events_batch(self) -> PayloadDictList:
        """
        Get a list of all the PayloadDicts in the buffer.

        :rtype  PayloadDictList
        """
        ...

    def cleanup(self, batch: PayloadDictList, need_retry: bool) -> None:
        """
        Removes sent events from the event store. If events need to be retried they are re-added to the buffer.

        :param  batch:  The events to be removed from the buffer
        :type   batch:  PayloadDictList
        :param  need_retry  Whether the events should be re-sent or not
        :type   need_retry  bool
        """
        ...

    def size(self) -> int:
        """
        Returns the number of events in the buffer

        :rtype  int
        """
        ...


class InMemoryEventStore(EventStore):
    """
    Create a InMemoryEventStore object with custom buffer capacity. The default is 10,000 events.
    """

    def __init__(self, logger: Logger, buffer_capacity: int = 10000) -> None:
        """
        :param  logger: Logging module
        :type   logger: Logger
        :param  buffer_capacity:    The maximum capacity of the event buffer.
                                    When the buffer is full new events are lost.
        :type   buffer_capacity     int
        """
        self.event_buffer: List[PayloadDict] = []
        self.buffer_capacity = buffer_capacity
        self.logger = logger

    def add_event(self, payload: PayloadDict) -> bool:
        """
        Add PayloadDict to buffer.

        :param payload: The payload to add
        :type  payload: PayloadDict
        """
        if self._buffer_capacity_reached():
            self.logger.error("Event buffer is full, dropping event.")
            return False

        self.event_buffer.append(payload)
        return True

    def get_events_batch(self) -> PayloadDictList:
        """
        Get a list of all the PayloadDicts in the in the buffer.

        :rtype  PayloadDictList
        """
        batch = self.event_buffer
        self.event_buffer = []
        return batch

    def cleanup(self, batch: PayloadDictList, need_retry: bool) -> None:
        """
        Removes sent events from the InMemoryEventStore buffer. If events need to be retried they are re-added to the buffer.

        :param  batch:  The events to be removed from the buffer
        :type   batch:  PayloadDictList
        :param  need_retry  Whether the events should be re-sent or not
        :type   need_retry  bool
        """
        if not need_retry:
            return

        for event in batch:
            if not event in self.event_buffer:
                if not self.add_event(event):
                    return

    def size(self) -> int:
        """
        Returns the number of events in the buffer

        :rtype  int
        """
        return len(self.event_buffer)

    def _buffer_capacity_reached(self) -> bool:
        """
        Returns true if buffer capacity is reached

        :rtype: bool
        """
        return self.size() >= self.buffer_capacity



# """
#     page_view.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """
from snowplow_tracker.events.event import Event
from typing import Optional, List
from snowplow_tracker.subject import Subject
from snowplow_tracker.self_describing_json import SelfDescribingJson
from snowplow_tracker.contracts import non_empty_string


class PageView(Event):
    """
    Constructs a PageView event object.

    When tracked, generates a "pv" or "page_view" event.

    """

    def __init__(
        self,
        page_url: str,
        page_title: Optional[str] = None,
        referrer: Optional[str] = None,
        event_subject: Optional[Subject] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        true_timestamp: Optional[float] = None,
    ) -> None:
        """
        :param  page_url:       URL of the viewed page
        :type   page_url:       non_empty_string
        :param  page_title:     Title of the viewed page
        :type   page_title:     string_or_none
        :param  referrer:       Referrer of the page
        :type   referrer:       string_or_none
        :param  event_subject:   Optional per event subject
        :type   event_subject:   subject | None
        :param  context:         Custom context for the event
        :type   context:         context_array | None
        :param  true_timestamp:          Optional event timestamp in milliseconds
        :type   true_timestamp:          int | float | None
        """
        super(PageView, self).__init__(
            event_subject=event_subject, context=context, true_timestamp=true_timestamp
        )
        self.payload.add("e", "pv")
        self.page_url = page_url
        self.page_title = page_title
        self.referrer = referrer

    @property
    def page_url(self) -> str:
        """
        URL of the viewed page
        """
        return self.payload.nv_pairs["url"]

    @page_url.setter
    def page_url(self, value: str):
        non_empty_string(value)
        self.payload.add("url", value)

    @property
    def page_title(self) -> Optional[str]:
        """
        Title of the viewed page
        """
        return self.payload.nv_pairs.get("page")

    @page_title.setter
    def page_title(self, value: Optional[str]):
        self.payload.add("page", value)

    @property
    def referrer(self) -> Optional[str]:
        """
        The referrer of the page
        """
        return self.payload.nv_pairs.get("refr")

    @referrer.setter
    def referrer(self, value: Optional[str]):
        self.payload.add("refr", value)



# """
#     struct_event.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """
from snowplow_tracker.events.event import Event
from typing import Optional, List, Union
from snowplow_tracker.subject import Subject
from snowplow_tracker.self_describing_json import SelfDescribingJson
from snowplow_tracker.contracts import non_empty_string


class StructuredEvent(Event):
    """
    Constructs a Structured event object.

    This event type is provided to be roughly equivalent to Google Analytics-style events.
    Note that it is not automatically clear what data should be placed in what field.
    To aid data quality and modeling, agree on business-wide definitions when designing
    your tracking strategy.

    We recommend using SelfDescribing - fully custom - events instead.

    When tracked, generates a "struct" or "se" event.
    """

    def __init__(
        self,
        category: str,
        action: str,
        label: Optional[str] = None,
        property_: Optional[str] = None,
        value: Optional[Union[int, float]] = None,
        event_subject: Optional[Subject] = None,
        context: Optional[List[SelfDescribingJson]] = None,
        true_timestamp: Optional[float] = None,
    ) -> None:
        """
        :param  category:       Category of the event
        :type   category:       non_empty_string
        :param  action:         The event itself
        :type   action:         non_empty_string
        :param  label:          Refer to the object the action is
                                performed on
        :type   label:          string_or_none
        :param  property_:      Property associated with either the action
                                or the object
        :type   property_:      string_or_none
        :param  value:          A value associated with the user action
        :type   value:          int | float | None
        :param  event_subject:   Optional per event subject
        :type   event_subject:   subject | None
        :param  context:         Custom context for the event
        :type   context:         context_array | None
        :param  true_timestamp:          Optional event timestamp in milliseconds
        :type   true_timestamp:          int | float | None
        """
        super(StructuredEvent, self).__init__(
            event_subject=event_subject, context=context, true_timestamp=true_timestamp
        )
        self.payload.add("e", "se")
        self.category = category
        self.action = action
        self.label = label
        self.property_ = property_
        self.value = value

    @property
    def category(self) -> Optional[str]:
        """
        Category of the event
        """
        return self.payload.nv_pairs.get("se_ca")

    @category.setter
    def category(self, value: str):
        non_empty_string(value)
        self.payload.add("se_ca", value)

    @property
    def action(self) -> Optional[str]:
        """
        The event itself
        """
        return self.payload.nv_pairs.get("se_ac")

    @action.setter
    def action(self, value: str):
        non_empty_string(value)
        self.payload.add("se_ac", value)

    @property
    def label(self) -> Optional[str]:
        """
        Refer to the object the action is performed on
        """
        return self.payload.nv_pairs.get("se_la")

    @label.setter
    def label(self, value: Optional[str]):
        self.payload.add("se_la", value)

    @property
    def property_(self) -> Optional[str]:
        """
        Property associated with either the action or the object
        """
        return self.payload.nv_pairs.get("se_pr")

    @property_.setter
    def property_(self, value: Optional[str]):
        self.payload.add("se_pr", value)

    @property
    def value(self) -> Optional[Union[int, float]]:
        """
        A value associated with the user action
        """
        return self.payload.nv_pairs.get("se_va")

    @value.setter
    def value(self, value: Optional[Union[int, float]]):
        self.payload.add("se_va", value)



#
#     setup.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

authors_list = [
    "Anuj More",
    "Alexander Dean",
    "Fred Blundun",
    "Paul Boocock",
    "Matus Tomlein",
    "Jack Keene",
]
authors_str = ", ".join(authors_list)

authors_email_list = [
    "support@snowplow.io",
]
authors_email_str = ", ".join(authors_email_list)

setup(
    name="snowplow-tracker",
    version="1.0.4",
    author=authors_str,
    author_email=authors_email_str,
    packages=["snowplow_tracker", "snowplow_tracker.test", "snowplow_tracker.events"],
    package_data={"snowplow_tracker": ["py.typed"]},
    url="http://snowplow.io",
    license="Apache License 2.0",
    description="Snowplow event tracker for Python. Add analytics to your Python and Django apps, webapps and games",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "requests>=2.25.1,<3.0",
        "types-requests>=2.25.1,<3.0",
        "typing_extensions>=3.7.4",
    ],
)



# """
#     contracts.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import traceback
import re
from typing import Any, Dict, Iterable, Callable, Sized
from snowplow_tracker.typing import FORM_TYPES, FORM_NODE_NAMES

_CONTRACTS_ENABLED = True
_MATCH_FIRST_PARAMETER_REGEX = re.compile(r"\(([\w.]+)[,)]")


def disable_contracts() -> None:
    global _CONTRACTS_ENABLED
    _CONTRACTS_ENABLED = False


def enable_contracts() -> None:
    global _CONTRACTS_ENABLED
    _CONTRACTS_ENABLED = True


def contracts_enabled() -> bool:
    global _CONTRACTS_ENABLED
    return _CONTRACTS_ENABLED


def greater_than(value: float, compared_to: float) -> None:
    if contracts_enabled() and value <= compared_to:
        raise ValueError(
            "{0} must be greater than {1}.".format(_get_parameter_name(), compared_to)
        )


def non_empty(seq: Sized) -> None:
    if contracts_enabled() and len(seq) == 0:
        raise ValueError("{0} is empty.".format(_get_parameter_name()))


def non_empty_string(s: str) -> None:
    if contracts_enabled() and type(s) is not str or not s:
        raise ValueError("{0} is empty.".format(_get_parameter_name()))


def one_of(value: Any, supported: Iterable) -> None:
    if contracts_enabled() and value not in supported:
        raise ValueError("{0} is not supported.".format(_get_parameter_name()))


def satisfies(value: Any, check: Callable[[Any], bool]) -> None:
    if contracts_enabled() and not check(value):
        raise ValueError("{0} is not allowed.".format(_get_parameter_name()))


def form_element(element: Dict[str, Any]) -> None:
    satisfies(element, lambda x: _check_form_element(x))


def _get_parameter_name() -> str:
    stack = traceback.extract_stack()
    _, _, _, code = stack[-3]

    match = _MATCH_FIRST_PARAMETER_REGEX.search(code)
    if not match:
        return "Unnamed parameter"
    return str(match.groups(0)[0])


def _check_form_element(element: Dict[str, Any]) -> bool:
    """
    Helper method to check that dictionary conforms element
    in sumbit_form and change_form schemas
    """
    all_present = (
        isinstance(element, dict)
        and "name" in element
        and "value" in element
        and "nodeName" in element
    )
    try:
        if element["type"] in FORM_TYPES:
            type_valid = True
        else:
            type_valid = False
    except KeyError:
        type_valid = True
    return all_present and element["nodeName"] in FORM_NODE_NAMES and type_valid



# """
#     tracker_configuration.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

from typing import Optional
from snowplow_tracker.typing import JsonEncoderFunction


class TrackerConfiguration(object):
    def __init__(
        self,
        encode_base64: bool = True,
        json_encoder: Optional[JsonEncoderFunction] = None,
    ) -> None:
        """
        Configuration for additional tracker configuration options.
        :param encode_base64:     Whether JSONs in the payload should be base-64 encoded. Default is True.
        :type  encode_base64:     bool
        :param json_encoder:      Custom JSON serializer that gets called on non-serializable object.
        :type  json_encoder:      function | None
        """

        self.encode_base64 = encode_base64
        self.json_encoder = json_encoder

    @property
    def encode_base64(self) -> bool:
        """
        Whether JSONs in the payload should be base-64 encoded. Default is True.
        """
        return self._encode_base64

    @encode_base64.setter
    def encode_base64(self, value: bool):
        if isinstance(value, bool) or value is None:
            self._encode_base64 = value

    @property
    def json_encoder(self) -> Optional[JsonEncoderFunction]:
        """
        Custom JSON serializer that gets called on non-serializable object.
        """
        return self._json_encoder

    @json_encoder.setter
    def json_encoder(self, value: Optional[JsonEncoderFunction]):
        self._json_encoder = value



# """
#     __init__.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """
from snowplow_tracker.events.event import Event
from snowplow_tracker.events.page_ping import PagePing
from snowplow_tracker.events.page_view import PageView
from snowplow_tracker.events.self_describing import SelfDescribing
from snowplow_tracker.events.structured_event import StructuredEvent
from snowplow_tracker.events.screen_view import ScreenView
